class State(object):
    def __init__(self, default=None):
        self.default = default
        self.commands = set()
        self.keys = {}
    
    def add(self, command):
        self.commands.add(command)
        for key in command.defaults:
            self.keys[key] = command
        return command
    
    def get(self, key):
        return self.keys.get(key, self.default)

base = State()

class Command(object):
    defaults = []
    next_state = base
    
    def run(self, world, window):
        pass

@base.add
class Quit(Command):
    # libtcod seems to return PRESSED for the Escape key.
    defaults = ["^q", "^c", "ESCAPE", "PRESSED"]
    next_state = None

class Reset(Command):
    def run(self, world, window):
        world.reset()

class Movement(Command):
    dx = dy = 0
    steps = 1
    
    def run(self, world, window):
        won = None
        x, y, z = world.player.loc
        for step in range(self.steps):
            if step:
                window.display(world)
            
            x += self.dx
            y += self.dy
            space = world[x,y,z]
            won = space.bump()
            if not space.passable:
                break
            z = space.passable
            world.player.move((x,y,z))
            if won:
                window.message("Level %d complete!" % sum(world.size))
                self.next_state = State(Reset)
                self.next_state.add(Quit)

@base.add
class MoveWest(Movement):
    defaults = ["h", "a", "KP4", "LEFT"]
    dx = -1
@base.add
class RunWest(MoveWest):
    defaults = ["H", "A", "S-KP4", "S-LEFT"]
    steps = 4

@base.add
class MoveSouth(Movement):
    defaults = ["j", "s", "x", "KP2", "DOWN"]
    dy = 1
@base.add
class RunSouth(MoveSouth):
    defaults = ["J", "S", "X", "S-KP2", "S-DOWN"]
    steps = 4

@base.add
class MoveNorth(Movement):
    defaults = ["k", "w", "KP8", "UP"]
    dy = -1
@base.add
class RunNorth(MoveNorth):
    defaults = ["K", "W", "S-KP8", "S-UP"]
    steps = 4

@base.add
class MoveEast(Movement):
    defaults = ["l", "d", "KP6", "RIGHT"]
    dx = 1
@base.add
class RunEast(MoveEast):
    defaults = ["L", "D", "S-KP6", "S-RIGHT"]
    steps = 4

@base.add
class MoveNorthwest(Movement):
    defaults = ["y", "q", "KP7", "HOME"]
    dx = -1
    dy = -1
@base.add
class RunNorthwest(MoveNorthwest):
    defaults = ["Y", "Q", "S-KP7", "S-HOME"]
    steps = 4

@base.add
class MoveNortheast(Movement):
    defaults = ["u", "e", "KP9", "PAGEUP"]
    dx = 1
    dy = -1
@base.add
class RunNortheast(MoveNortheast):
    defaults = ["U", "E", "S-KP9", "S-PAGEUP"]
    steps = 4

@base.add
class MoveSouthwest(Movement):
    defaults = ["b", "z", "KP1", "END"]
    dx = -1
    dy = 1
@base.add
class RunSouthwest(MoveSouthwest):
    defaults = ["B", "Z", "S-KP1", "S-END"]
    steps = 4

@base.add
class MoveSoutheast(Movement):
    defaults = ["n", "c", "KP3", "PAGEDOWN"]
    dx = 1
    dy = 1
@base.add
class RunSoutheast(MoveSoutheast):
    defaults = ["N", "C", "S-KP3", "S-PAGEDOWN"]
    steps = 4
